import numpy as np
import pandas as pd

from .utils import get_data_series, group_yearly_df, validate_series


def rai(series: pd.Series) -> pd.Series:
    """
    Calculate the Rainfall Anomaly Index (RAI) for a given time
    series of precipitation data. [vanrooy_1965]_

    Parameters
    ----------
    series : pd.Series
        A pandas Series containing precipitation data.

    Returns
    -------
    pd.Series
        A pandas Series containing the RAI values.

    References
    ----------
    .. [vanrooy_1965] van Rooy, M.P. (1965). A Rainfall Anomaly
       Index Independent of Time and Space. Notos.
    """
    series = validate_series(series)
    pm = series.mean()
    pi_above = series > pm
    rai = pd.Series(np.nan, index=series.index, dtype=float)
    rai[pi_above] = 3.0 * (series[pi_above] - pm) / (series.nlargest(10).mean() - pm)
    rai[~pi_above] = (
        -3.0 * (series[~pi_above] - pm) / (series.nsmallest(10).mean() - pm)
    )
    return rai


def mrai(series: pd.Series, sf: float = 1.7) -> pd.Series:
    """Calculate the Modified Rainfall Anomaly Index (MRAI) for a
    given time series of precipitation data. [hansel_2015]_

    Parameters
    ----------
    series : pd.Series
        A pandas Series containing precipitation data.
    sf : float
        Scaling factor for the MRAI calculation. Default is 1.7.

    Returns
    -------
    pd.Series
        A pandas Series containing the MRAI values.

    References
    ----------
    .. [hansel_2015] Hänsel, S., Schucknecht, A. and Matschullat J. (2015).
       The Modified Rainfall Anomaly Index (mRAI) — is this an alternative
       to the Standardised Precipitation Index (SPI) in evaluating future
       extreme precipitation characteristics? Theoretical and Applied
       Climatology. doi.org/10.1007/s00704-015-1389-y.
    """
    series = validate_series(series)
    mrai = pd.Series(np.nan, index=series.index, dtype=float)
    group_df = group_yearly_df(series=series)
    for _, gr in group_df.groupby(pd.Grouper(freq="MS")):
        gr_series = get_data_series(gr)
        pm = gr_series.mean()
        pi_above = gr_series > pm
        e_above = gr_series[gr_series > gr_series.quantile(0.9)].mean()
        e_below = gr_series[gr_series < gr_series.quantile(0.1)].mean()
        mrai_gr = pd.Series(np.nan, index=gr_series.index, dtype=float)
        mrai_gr[pi_above] = sf * (gr_series[pi_above] - pm) / (e_above - pm)
        mrai_gr[~pi_above] = -sf * (gr_series[~pi_above] - pm) / (e_below - pm)
        mrai.loc[mrai_gr.index] = mrai_gr.values

    return mrai



import logging
from dataclasses import dataclass, field
from typing import Literal

from numpy import ceil, interp, linspace, nan
from pandas import DataFrame, Grouper, Series, Timedelta, Timestamp
from scipy.stats import beta, fisk, gamma, genextreme, norm

from ._typing import ContinuousDist
from .dist import Dist
from .utils import (
    daily_window_group_yearly_df,
    get_data_series,
    group_yearly_df,
    infer_frequency,
    validate_series,
)


def sgi(
    series: Series,
    timescale: int = 0,
    fit_freq: str | None = None,
) -> Series:
    """Method to compute the Standardized Groundwater Index [sgi_2013]_. Same
    method as in Pastas. Uses the normal scores transform to calculate the
    cumulative density function.

    Parameters
    ----------
    series: pandas.Series
        Pandas time series of the groundwater levels. Time series index
        should be a pandas DatetimeIndex.
    fit_freq : str, optional, default=None
        Frequency for fitting the distribution. Default is None in which case
        the frequency of the series is inferred. If this fails a monthly
        frequency is used.

    Returns
    -------
    pandas.Series

    References
    ----------
    .. [sgi_2013] Bloomfield, J. P. and Marchant, B. P.: Analysis of
       groundwater drought building on the standardised precipitation index
       approach. Hydrol. Earth Syst. Sci., 17, 4769–4787, 2013.
    """

    mock_dist = norm
    sgi = SI(
        series=series,
        dist=mock_dist,
        timescale=timescale,
        fit_freq=fit_freq,
        fit_window=0,
        prob_zero=False,
        normal_scores_transform=True,
        agg_func="mean",
    )
    return sgi.norm_ppf()


def spi(
    series: Series,
    dist: ContinuousDist = gamma,
    timescale: int = 0,
    fit_freq: str | None = None,
    fit_window: int = 0,
    prob_zero: bool = True,
) -> Series:
    """Method to compute the Standardized Precipitation Index [spi_2002]_.

    Parameters
    ----------
    series: pandas.Series
        Pandas time series of the precipitation. Time series index
        should be a pandas DatetimeIndex.
    dist: scipy.stats.rv_continuous
        Can be any continuous distribution from the scipy.stats library.
        However, for the SPI generally the Gamma probability density
        function is recommended. Other appropriate choices could be the
        lognormal, log-logistic (fisk) or PearsonIII distribution.
    timescale : int, optional, default=0
        Size of the moving window over which the series is summed. If zero, no
        summation is performed over the time series. If the time series
        frequency is daily, then one would provide timescale=30 for SI1,
        timescale=90 for SI3, timescale=180 for SI6 etc.
    fit_freq : str, optional, default=None
        Frequency for fitting the distribution. Default is None in which case
        the frequency of the series is inferred. If this fails a monthly
        frequency is used.
    fit_window : int, optional, default=0
        Window size for fitting data in fit_freq frequency's unit. Default is
        zero in which case only data within the fit_freq is considered. If
        larger than zero data data within the window is used to fit the
        distribution for the series. fit_window must be a odd number larger
        than 3 when used.
    prob_zero : bool, default=True
        Option to correct the distribution if x=0 is not in probability density
        function. E.g. the case with the Gamma distriubtion. If True, the
        probability of zero values in the series is calculated by the
        occurence.

    Returns
    -------
    pandas.Series

    References
    ----------
    .. [spi_2002] LLoyd-Hughes, B. and Saunders, M.A.: A drought
       climatology for Europe. International Journal of Climatology,
       22, 1571-1592, 2002.
    """

    spi = SI(
        series=series,
        dist=dist,
        timescale=timescale,
        fit_freq=fit_freq,
        fit_window=fit_window,
        prob_zero=prob_zero,
        normal_scores_transform=False,
        agg_func="sum",
    )
    spi.fit_distribution()
    return spi.norm_ppf()


def spei(
    series: Series,
    dist: ContinuousDist = fisk,
    timescale: int = 0,
    fit_freq: str | None = None,
    fit_window: int = 0,
    prob_zero: bool = False,
) -> Series:
    """Method to compute the Standardized Precipitation Evaporation Index
    [spei_2010]_.

    Parameters
    ----------
    series: pandas.Series
        Pandas time series of the precipitation. Time series index
        should be a pandas DatetimeIndex.
    dist: scipy.stats.rv_continuous
        Can be any continuous distribution from the scipy.stats library.
        However, for the SPEI generally the log-logistic (fisk) probability
        density function is recommended. Other appropriate choices could be
        the lognormal or PearsonIII distribution.
    timescale : int, optional, default=0
        Size of the moving window over which the series is summed. If zero, no
        summation is performed over the time series. If the time series
        frequency is daily, then one would provide timescale=30 for SI1,
        timescale=90 for SI3, timescale=180 for SI6 etc.
    fit_freq : str, optional, default=None
        Frequency for fitting the distribution. Default is None in which case
        the frequency of the series is inferred. If this fails a monthly
        frequency is used.
    fit_window : int, optional, default=0
        Window size for fitting data in fit_freq frequency's unit. Default is
        zero in which case only data within the fit_freq is considered. If
        larger than zero data data within the window is used to fit the
        distribution for the series. fit_window must be a odd number larger
        than 3 when used.
    prob_zero : bool, default=False
        Flag indicating whether the probability of zero values in the series is
        calculated by the occurence.

    Returns
    -------
    pandas.Series

    References
    ----------
    .. [spei_2010] Vicente-Serrano S.M., Beguería S., López-Moreno J.I.:
       A Multi-scalar drought index sensitive to global warming: The
       Standardized Precipitation Evapotranspiration Index.
       Journal of Climate, 23, 1696-1718, 2010.
    """

    spei = SI(
        series=series,
        dist=dist,
        timescale=timescale,
        fit_freq=fit_freq,
        fit_window=fit_window,
        prob_zero=prob_zero,
        normal_scores_transform=False,
        agg_func="sum",
    )
    spei.fit_distribution()
    return spei.norm_ppf()


def ssfi(
    series: Series,
    dist: ContinuousDist = genextreme,
    timescale: int = 0,
    fit_freq: str | None = None,
    fit_window: int = 0,
    prob_zero: bool = True,
) -> Series:
    """Method to compute the Standardized StreamFlow Index [ssfi_2012]_.

    Parameters
    ----------
    series: pandas.Series
        Pandas time series of the precipitation. Time series index
        should be a pandas DatetimeIndex.
    dist: scipy.stats.rv_continuous
        Can be any continuous distribution from the scipy.stats library.
        However, for the SSFI generally the gamma probability density function
        is recommended. Other choices could be the normal, lognormal,
        pearsonIII, GEV or Gen-Logistic distribution or any distribution deemed
        appropriate.
    timescale : int, optional, default=0
        Size of the moving window over which the series is summed. If zero, no
        summation is performed over the time series. If the time series
        frequency is daily, then one would provide timescale=30 for SI1,
        timescale=90 for SI3, timescale=180 for SI6 etc.
    fit_freq : str, optional, default=None
        Frequency for fitting the distribution. Default is None in which case
        the frequency of the series is inferred. If this fails a monthly
        frequency is used.
    fit_window : int, optional, default=0
        Window size for fitting data in fit_freq frequency's unit. Default is
        zero in which case only data within the fit_freq is considered. If
        larger than zero data data within the window is used to fit the
        distribution for the series. fit_window must be a odd number larger
        than 3 when used.
    prob_zero : bool, default=False
        Flag indicating whether the probability of zero values in the series is
        calculated by the occurence.

    Returns
    -------
    pandas.Series

    References
    ----------
    .. [ssfi_2012] Vicente-Serrano, S. M., J. I. López-Moreno, S. Beguería,
    J. Lorenzo-Lacruz, C. Azorin-Molina, and E. Morán-Tejeda (2012). Accurate
    Computation of a Streamflow Drought Index. Journal of Hydrologic
    Engineering 17 (2): 318-332.
    """
    ssfi = SI(
        series=series,
        dist=dist,
        timescale=timescale,
        fit_freq=fit_freq,
        fit_window=fit_window,
        prob_zero=prob_zero,
        normal_scores_transform=False,
        agg_func="mean",
    )
    ssfi.fit_distribution()
    return ssfi.norm_ppf()


def ssmi(
    series: Series,
    dist: ContinuousDist = beta,
    timescale: int = 0,
    fit_freq: str | None = None,
    fit_window: int = 0,
    prob_zero: bool = True,
) -> Series:
    """Method to compute the Standardized Soil Moisture Index [ssmi_2016]_.

    Parameters
    ----------
    series: pandas.Series
        Pandas time series of the precipitation. Time series index
        should be a pandas DatetimeIndex.
    dist: scipy.stats.rv_continuous
        Can be any continuous distribution from the scipy.stats library.
        However, for the SSMI generally the beta probability density function
        is recommended. Other choices could be the normal or ECDF distribution
        or any distribution deemed appropriate.
    timescale : int, optional, default=0
        Size of the moving window over which the series is summed. If zero, no
        summation is performed over the time series. If the time series
        frequency is daily, then one would provide timescale=30 for SI1,
        timescale=90 for SI3, timescale=180 for SI6 etc.
    fit_freq : str, optional, default=None
        Frequency for fitting the distribution. Default is None in which case
        the frequency of the series is inferred. If this fails a monthly
        frequency is used.
    fit_window : int, optional, default=0
        Window size for fitting data in fit_freq frequency's unit. Default is
        zero in which case only data within the fit_freq is considered. If
        larger than zero data data within the window is used to fit the
        distribution for the series. fit_window must be a odd number larger
        than 3 when used.
    prob_zero : bool, default=False
        Flag indicating whether the probability of zero values in the series is
        calculated by the occurence.

    Returns
    -------
    pandas.Series

    References
    ----------
    .. [ssmi_2016] Carrão. H., Russo, S., Sepulcre-Canto, G., Barbosa, P.: An
       empirical standardized soil moisture index for agricultural drought assessment
       from remotely sensed data. International Journal of Applied Earth Observation
       and Geoinformation, 48, 2016.
    """

    ssmi = SI(
        series=series,
        dist=dist,
        timescale=timescale,
        fit_freq=fit_freq,
        fit_window=fit_window,
        prob_zero=prob_zero,
        normal_scores_transform=False,
        agg_func="mean",
    )
    ssmi.fit_distribution()
    return ssmi.norm_ppf()


@dataclass
class SI:
    """
    Standardized Index Class.

    Parameters
    ----------
    series : Series
        The input time series data.
    dist : ContinuousDist
        The SciPy continuous distribution associated with the data.
    timescale : int, optional, default=0
        Size of the moving window over which the series is summed. If zero, no
        summation is performed over the time series. If the time series
        frequency is daily, then one would provide timescale=30 for SI1,
        timescale=90 for SI3, timescale=180 for SI6 etc.
    fit_freq : str, optional, default=None
        Frequency for fitting the distribution. Default is None in which case
        the frequency of the series is inferred. If this fails a monthly
        frequency is used.
    fit_window : int, optional, default=0
        Window size for fitting data in fit_freq frequency's unit. Default is
        zero in which case only data within the fit_freq is considered. If
        larger than zero data data within the window is used to fit the
        distribution for the series. fit_window must be a odd number larger
        than 3 when used.
    prob_zero : bool, default=False
        Flag indicating whether the probability of zero values in the series is
        calculated by the occurence.
    normal_scores_transform : bool, default=False
        Flag to use the normal scores transformation for calculating the
        cumulative density function.
    agg_func: Literal['sum', 'mean'], default='sum'
        String of the function to use for aggregating the time series if the
        timescale is larger than 0. Can either be 'sum' or 'mean'.

    Attributes
    ----------
    _grouped_year : DataFrame
        Dataframe with all data grouped in a one-year (2000) DataFrame with the
        original years as columns
    _dist_dict : Dict[int, Dist]
        Dictionary of distributions used to fit the data.
    """

    series: Series = field(repr=False)
    dist: ContinuousDist
    timescale: int = 0
    fit_freq: str | None = field(default=None)
    fit_window: int = field(default=0)
    prob_zero: bool = field(default=False)
    normal_scores_transform: bool = field(default=False)
    agg_func: Literal["sum", "mean"] = "sum"
    _grouped_year: DataFrame = field(init=False, repr=False, compare=False)
    _dist_dict: dict[int, Dist] = field(
        default_factory=dict, init=False, repr=False, compare=False
    )

    def __post_init__(self) -> None:
        """
        Post initializes the SI class and performs necessary data
        preprocessing and validation.
        """
        self.series = validate_series(self.series)

        if self.timescale > 0:
            self.series = (
                self.series.rolling(self.timescale, min_periods=self.timescale)
                .agg(self.agg_func)
                .dropna()
                .copy()
            )

        if self.fit_freq is None:
            self.fit_freq = infer_frequency(self.series.index)

        self._grouped_year = group_yearly_df(series=self.series)

        if self.fit_window > 0:
            if self.fit_window < 3:
                logging.error(
                    "Window should be larger than 2. Setting the window value to 3."
                )
                self.fit_window = 3  # make sure window is at least three
            elif self.fit_window % 2 == 0:
                logging.error(
                    "Window should be odd. Setting the window value to"
                    f"{self.fit_window + 1}"
                )
                self.fit_window += 1  # make sure window is odd

    def fit_distribution(self) -> None:
        """
        Fit distribution on the time series per fit_frequency and/or fit_window
        """

        if self.normal_scores_transform:
            logging.info("Using normal-scores-transform. No distribution is fitted.")

        elif self.fit_window > 0:
            if self.fit_freq not in (
                "d",
                "w",
                "D",
                "W",
            ):  # TODO: ideally 14D should also work.
                raise ValueError(
                    "Frequency fit_freq must be 'D' or 'W', not "
                    f"'{self.fit_freq}', if a fit_window is provided."
                )

            logging.info("Using rolling window method")
            window = self.fit_window
            period = int(ceil(window / 2))
            if self.fit_freq in ("W", "w"):
                period = Timedelta(value=period, unit="W").days
                window = period * 2 + 1

            dfval_window = daily_window_group_yearly_df(
                dfval=self._grouped_year, period=period
            )
            for dfval_rwindow in dfval_window.rolling(
                window=window, min_periods=window, closed="right"
            ):
                if len(dfval_rwindow) < window:
                    continue  # min_periods ignored by Rolling.__iter__
                date = dfval_rwindow.index[period]
                data = get_data_series(dfval_rwindow.loc[[date]])
                data_window = get_data_series(dfval_rwindow)
                fd = Dist(
                    data=data,
                    dist=self.dist,
                    prob_zero=self.prob_zero,
                    data_window=data_window,
                )
                self._dist_dict[date] = fd
        else:
            logging.info("Using groupby fit by frequency method")
            for date, grval in self._grouped_year.groupby(
                Grouper(freq=str(self.fit_freq))
            ):
                data = get_data_series(grval)
                fd = Dist(
                    data=data,
                    dist=self.dist,
                    prob_zero=self.prob_zero,
                    data_window=None,
                )
                self._dist_dict[date] = fd

    def cdf(self) -> Series:
        """Compute the cumulative density function"""
        if self.normal_scores_transform:
            cdf = self.cdf_nsf()
        else:
            cdf = Series(nan, index=self.series.index, dtype=float)
            for k in self._dist_dict:
                cdf_k = self._dist_dict[k].cdf()
                cdf.loc[cdf_k.index] = cdf_k.values

        return cdf

    def pdf(self) -> Series:
        """Compute the probability density function"""
        if self.normal_scores_transform:
            pdf = self.cdf().diff()
        else:
            pdf = Series(nan, index=self.series.index, dtype=float)
            for k in self._dist_dict:
                pdf_k = self._dist_dict[k].pdf()
                pdf.loc[pdf_k.index] = pdf_k.values
        return pdf

    def cdf_nsf(self) -> Series:
        """
        Compute the cumulative density function using the Normal Scores
        Transform

        Returns
        -------
        Series
        """
        logging.info("Using the normal scores transform")
        cdf = Series(nan, index=self.series.index, dtype=float)
        for _, grval in self._grouped_year.groupby(Grouper(freq=str(self.fit_freq))):
            data = get_data_series(grval).sort_values()
            n = len(data)
            cdf.loc[data.index] = linspace(1 / (2 * n), 1 - 1 / (2 * n), n)
        return cdf

    def ppf(self, q: float) -> Series:
        """
        Method to calculate the percentile point function
        (inverse of cdf — percentiles) of a fitted
        distribution.

        Parameters
        ----------
        q : float
            The quantile value (between 0 and 1) for which to calculate the
            percentile point function.

        Returns
        -------
        Series
        """
        ppf = Series(nan, index=self.series.index, dtype=float)
        if self.normal_scores_transform:
            cdf = self.cdf_nsf()
            for _, grval in self._grouped_year.groupby(
                Grouper(freq=str(self.fit_freq))
            ):
                data = get_data_series(grval).sort_values()
                cdf_i = cdf.loc[data.index]
                ppf.loc[data.index] = interp(
                    x=q,
                    xp=cdf_i.values.astype(float),
                    fp=data.values.astype(float),
                )
        else:
            for k in self._dist_dict:
                ppf_k = self._dist_dict[k].ppf(q=q)
                ppf.loc[ppf_k.index] = ppf_k.values
        return ppf

    def norm_ppf(self) -> Series:
        """
        Method to calculate propability point function of normal distribution
        based on a cumulative density function of a fitted distribution

        Returns
        -------
        Series
        """

        cdf = self.cdf()
        ppf = Series(
            norm.ppf(cdf.values, loc=0, scale=1), index=self.series.index, dtype=float
        )
        return ppf

    def get_dist(self, date: Timestamp) -> Dist:
        for k in self._dist_dict:
            dist = self._dist_dict[k]
            if date in dist.data.index:
                return dist

        raise KeyError("Date not found in distributions")



import matplotlib as mpl
from pandas import Series

from spei.plot import heatmap, monthly_density, threshold
from spei.plot import si as plot_si

mpl.use("Agg")  # prevent _tkinter.TclError: Can't find a usable tk.tcl error


def test_plot_si(si: Series) -> None:
    _ = plot_si(si)


def test_plot_si_no_background(si: Series) -> None:
    _ = plot_si(si, cmap="roma_r", background=False)


def test_plot_monthly_density(si: Series) -> None:
    _ = monthly_density(si, years=[2011], months=[1, 2, 3, 4, 5])


def test_plot_heatmap(si: Series) -> None:
    _ = heatmap([si], cmap="vik", vmin=-3.0, vmax=3.0)


def test_plot_threshold(head: Series) -> None:
    th = Series(head.mean(), index=head.index, dtype=float)
    _ = threshold(series=head, threshold=th, fill_color="orange")



from pandas import DataFrame, Series, Timestamp
from scipy.stats import norm

from spei import SI, sgi, spei, spi, ssfi, ssmi
from spei.dist import Dist


def test_spi(prec: Series) -> None:
    precr = prec.rolling("30D", min_periods=30).sum().dropna()
    spi(precr, fit_freq="MS", prob_zero=True)


def test_spei(prec: Series, evap: Series) -> None:
    n = (prec - evap).rolling("30D", min_periods=30).sum().dropna()
    spei(n, fit_freq="MS")


def test_sgi(head: Series) -> None:
    sgi(head, fit_freq="MS")


def test_sffi_timescale(prec: Series) -> None:
    ssfi(prec, timescale=30)


def test_ssmi(prec: Series) -> None:
    ssmi(prec, fit_freq="MS")


def test_window(prec: Series, evap: Series) -> None:
    n = (prec - evap).rolling("30D", min_periods=30).sum().dropna()
    spei(n, fit_freq="W", fit_window=3)


def test_window_even(prec: Series, evap: Series, caplog) -> None:
    n = (prec - evap).rolling("30D", min_periods=30).sum().dropna()
    spei(n, fit_freq="W", fit_window=4)
    assert "Window should be odd. Setting the window value to" in caplog.text


def test_SI(prec: Series) -> None:
    si = SI(prec, dist=norm, timescale=30, fit_freq="MS")
    si.fit_distribution()
    si.pdf()
    dist = si.get_dist(Timestamp("2010-01-01"))
    dist.ks_test()


def test_SI_post_init_timescale(prec: Series) -> None:
    si = SI(prec, dist=norm, timescale=30, fit_freq="MS")
    assert si.series.equals(prec.rolling(30, min_periods=30).sum().dropna()), (
        "Timescale rolling sum not applied correctly"
    )


def test_SI_post_init_fit_freq_infer(prec: Series) -> None:
    si = SI(prec, dist=norm, timescale=0)
    assert si.fit_freq is not None, "Frequency inference failed"


def test_SI_post_init_grouped_year(prec: Series) -> None:
    si = SI(prec, dist=norm, timescale=0, fit_freq="MS")
    assert isinstance(si._grouped_year, DataFrame), "Grouped year DataFrame not created"


def test_SI_post_init_fit_window_adjustment(prec: Series) -> None:
    si = SI(prec, dist=norm, timescale=0, fit_freq="D", fit_window=2)
    assert si.fit_window == 3, "Fit window not adjusted to odd number"


def test_SI_post_init_fit_window_minimum(prec: Series) -> None:
    si = SI(prec, dist=norm, timescale=0, fit_freq="D", fit_window=1)
    assert si.fit_window == 3, "Fit window not adjusted to minimum value"


def test_fit_distribution_normal_scores_transform(prec: Series) -> None:
    si = SI(prec, dist=norm, timescale=30, fit_freq="MS", normal_scores_transform=True)
    si.fit_distribution()
    assert not si._dist_dict, (
        "Distribution dictionary should be empty when using normal scores transform"
    )


def test_fit_distribution_with_fit_window(prec: Series) -> None:
    si = SI(prec, dist=norm, timescale=30, fit_freq="D", fit_window=5)
    si.fit_distribution()
    assert si._dist_dict, (
        "Distribution dictionary should not be empty when using fit window"
    )
    for dist in si._dist_dict.values():
        assert isinstance(dist, Dist), (
            "Items in distribution dictionary should be of type Dist"
        )


def test_fit_distribution_with_fit_freq(prec: Series) -> None:
    si = SI(prec, dist=norm, timescale=30, fit_freq="MS")
    si.fit_distribution()
    assert si._dist_dict, (
        "Distribution dictionary should not be empty when using fit frequency"
    )
    for dist in si._dist_dict.values():
        assert isinstance(dist, Dist), (
            "Items in distribution dictionary should be of type Dist"
        )


def test_fit_distribution_invalid_fit_freq_with_window(prec: Series) -> None:
    si = SI(prec, dist=norm, timescale=30, fit_freq="M", fit_window=5)
    try:
        si.fit_distribution()
    except ValueError as e:
        assert (
            str(e)
            == "Frequency fit_freq must be 'D' or 'W', not 'M', if a fit_window is provided."
        )
    else:
        assert False, "ValueError not raised for invalid fit frequency with fit window"


def test_ppf(prec: Series) -> None:
    si = SI(prec, dist=norm, timescale=1, fit_freq="MS")
    si.fit_distribution()
    ppf = si.ppf(0.5)
    assert isinstance(ppf, Series), "PPF result should be a Pandas Series"
    assert len(ppf) == len(si.series), (
        "PPF result length does not match input series length"
    )


def test_ppf_nsf(prec: Series) -> None:
    si = SI(prec, dist=norm, timescale=1, fit_freq="MS", normal_scores_transform=True)
    si.fit_distribution()
    ppf = si.ppf(0.5)
    assert isinstance(ppf, Series), "PPF result should be a Pandas Series"
    assert len(ppf) == len(si.series), (
        "PPF result length does not match input series length"
    )



import numpy as np
import pandas as pd

from .utils import group_yearly_df, validate_series


def get_yearly_temp_date(temp: pd.Series, threshold: float) -> pd.Series:
    """
    Get the first date in each year where the cumulative temperature exceeds a given threshold.

    Parameters
    ----------
    temp : pd.Series
        A pandas Series representing the temperature time series, indexed by date.
    threshold : float
        The temperature threshold to identify the first date above it.

    Returns
    -------
    pd.Series
        A pandas Series containing the first date in each year where the cumulative
        temperature exceeds the threshold. The index corresponds to the years.
    """
    temp_group_df = group_yearly_df(series=temp).cumsum(axis=0)
    first_date_above_threshold = temp_group_df.gt(threshold).idxmax()
    return first_date_above_threshold


def cumsum(deficit: pd.Series, allow_below_zero: bool = True) -> pd.Series:
    """
    Calculate the cumulative sum of a deficit series.

    Parameters:
    -----------
    deficit : pd.Series
        A pandas Series representing the deficit values.
    allow_below_zero : bool, optional
        If True, the cumulative sum is calculated as-is, allowing negative values.
        If False, the cumulative sum is constrained to be non-negative, resetting
        to zero whenever the sum would drop below zero. Default is True.

    Returns:
    --------
    pd.Series
        A pandas Series containing the cumulative sum of the deficit values,
        optionally constrained to be non-negative.
    """
    if allow_below_zero:
        return deficit.cumsum()
    else:
        if deficit.iat[0] < 0.0:
            deficit.iat[0] = 0.0
        sumlm = np.frompyfunc(lambda a, b: 0.0 if a + b < 0.0 else a + b, nin=2, nout=1)
        return pd.Series(sumlm.accumulate(deficit.values), deficit.index, dtype=float)


def get_cumulative_deficit(
    deficit: pd.Series,
    startdate: pd.Timestamp | pd.Series,
    enddate: pd.Timestamp | pd.Series,
    allow_below_zero: bool = True,
) -> pd.DataFrame:
    """
    Calculate the cumulative deficit for a given time period.

    This function computes the cumulative deficit for each column in a
    grouped yearly DataFrame, starting from `startdate` to `enddate`.
    The cumulative sum can optionally allow values below zero.

    Parameters:
    -----------
    deficit : pd.Series
        A pandas Series representing the deficit time series.
    startdate : pd.Timestamp | pd.Series
        The start date(s) for the cumulative deficit calculation. If a
        single timestamp is provided, it is applied to all columns. If
        a Series is provided, it should align with the columns of the
        grouped DataFrame.
    enddate : pd.Timestamp | pd.Series
        The end date(s) for the cumulative deficit calculation. Similar
        to `startdate`, it can be a single timestamp or a Series aligned
        with the columns.
    allow_below_zero : bool, optional
        If True, allows the cumulative sum to include values below zero.
        Defaults to True.

    Returns:
    --------
    pd.DataFrame
        A DataFrame containing the cumulative deficit for each column
        over the specified time period. The index represents the date
        range, and the columns correspond to the year.
    """
    deficit = validate_series(deficit)
    group_df = group_yearly_df(series=deficit)
    if isinstance(startdate, pd.Timestamp):
        if startdate.year != 2000:
            startdate = startdate.replace(year=2000)
        startdate = pd.Series(startdate, index=group_df.columns)
    if isinstance(enddate, pd.Timestamp):
        if enddate.year != 2000:
            enddate = enddate.replace(year=2000)
        enddate = pd.Series(enddate, index=group_df.columns)

    index = pd.date_range(start=startdate.min(), end=enddate.max(), freq="D")
    cumdf = pd.DataFrame(np.nan, index=index, columns=group_df.columns)
    for col in group_df.columns:
        start = startdate[col]
        end = enddate[col]
        cumdf.loc[start:end, col] = cumsum(
            group_df.loc[start:end, col],
            allow_below_zero=allow_below_zero,
        ).values

    return cumdf


def deficit_oct1(deficit: pd.Series) -> pd.Series:
    """
    Calculate the cumulative deficit on October 1st.

    This function computes the cumulative deficit for a given time series
    of deficits, considering only the period between April 1st and
    September 30th. The cumulative deficit is reset to zero if it goes
    below zero during this period.

    Parameters:
    -----------
    deficit : pd.Series
        A pandas Series representing the deficit time series. The index
        should be datetime-like, and the values should represent the
        deficit amounts.

    Returns:
    --------
    pd.Series
        A pandas Series containing the cumulative deficit values on
        October 1st. The index of the returned Series corresponds to
        the columns of the cumulative deficit DataFrame, and the name
        of the Series is "Doct1".
    """
    startdate = pd.Timestamp("2000-04-01")
    enddate = pd.Timestamp("2000-09-30")
    cumdf = get_cumulative_deficit(
        deficit=deficit,
        startdate=startdate,
        enddate=enddate,
        allow_below_zero=False,
    )
    doct1 = pd.Series(
        data=cumdf.loc[enddate].values,
        index=cumdf.columns,
        dtype=float,
        name="Doct1",
    )
    return doct1


def deficit_max(deficit: pd.Series) -> pd.Series:
    """
    Calculate the maximum cumulative deficit within a specified period.

    This function computes the maximum cumulative deficit for a given
    deficit time series, starting from April 1st to September 30th.
    The cumulative deficit is calculated using the `get_cumulative_deficit`
    function, ensuring that values below zero are not allowed.

    Parameters:
    ----------
    deficit : pd.Series
        A pandas Series representing the deficit values over time.

    Returns:
    -------
    pd.Series
        A pandas Series containing the maximum cumulative deficit
        within the specified period, labeled as "Dmax".
    """
    startdate = pd.Timestamp("2000-04-01")
    enddate = pd.Timestamp("2000-09-30")
    cumdf = get_cumulative_deficit(
        deficit=deficit,
        startdate=startdate,
        enddate=enddate,
        allow_below_zero=False,
    )
    return cumdf.max().rename("Dmax")


def deficit_apr1(deficit: pd.Series) -> pd.Series:
    """
    Calculate the maximum change in cumulative deficit within a specified date range.

    This function computes the cumulative deficit for the given deficit series
    between April 1st and September 30th of the year 2000. It then calculates
    the maximum change in the cumulative deficit over this period.

    Parameters
    ----------
    deficit : pd.Series
        A pandas Series representing the deficit values. The index is expected
        to be datetime-like.

    Returns
    -------
    pd.Series
        A pandas Series containing the maximum change in cumulative deficit
        over the specified period, labeled as "DIapr1".
    """
    startdate = pd.Timestamp("2000-04-01")
    enddate = pd.Timestamp("2000-09-30")
    cumdf = get_cumulative_deficit(
        deficit=deficit,
        startdate=startdate,
        enddate=enddate,
        allow_below_zero=True,
    )
    return (cumdf.max() - cumdf.min()).rename("DIapr1")


def deficit_gdd(
    deficit: pd.Series, temp: pd.Series, threshold: float = 440.0
) -> pd.Series:
    """
    Calculate the maximum change in cumulative deficit starting from the
    first day when the temperature sum (growing degree days; GDD)
    exceeds a specified threshold.

    Parameters:
    -----------
    deficit : pd.Series
        A pandas Series representing the daily deficit values.
    temp : pd.Series
        A pandas Series representing the daily temperature values.
    threshold : float, optional
        The temperature sum GDD threshold to determine the starting date for the calculation.
        Defaults to 440.0.

    Returns:
    --------
    pd.Series
        A pandas Series containing the maximum change in cumulative deficit,
        labeled as "DIgdd".
    """
    temp = validate_series(temp)
    startdate = get_yearly_temp_date(temp=temp, threshold=threshold)
    enddate = pd.Timestamp("2000-09-30")
    cumdf = get_cumulative_deficit(
        deficit=deficit,
        startdate=startdate,
        enddate=enddate,
        allow_below_zero=True,
    )
    return (cumdf.max() - cumdf.min()).rename("DIapr1").rename("DIgdd")


def deficit_wet(deficit: pd.Series) -> pd.Series:
    """
    Calculate the maximum change in cumulative deficit for a specified period.

    This function computes the maximum change in  cumulative deficit from
    January 1st to September 30th of a given year. The cumulative deficit
    is calculated using the `get_cumulative_deficit` function, allowing
    values below zero.

    Parameters:
    -----------
    deficit : pd.Series
        A pandas Series representing the deficit values over time.

    Returns:
    --------
    pd.Series
        A pandas Series containing the maximum change in cumulative deficit
        for the specified period, labeled as "DIwet".
    """
    startdate = pd.Timestamp("2000-01-01")
    enddate = pd.Timestamp("2000-09-30")
    cumdf = get_cumulative_deficit(
        deficit=deficit,
        startdate=startdate,
        enddate=enddate,
        allow_below_zero=True,
    )
    return (cumdf.max() - cumdf.min()).rename("DIapr1").rename("DIwet")



from pandas import Series

from spei.rai import mrai, rai


def test_rai(prec: Series) -> None:
    precrs = prec.resample("MS").sum()
    rai_result = rai(precrs)
    assert isinstance(rai_result, Series), "RAI result is not a pandas Series"
    assert len(rai_result) == len(precrs), (
        "RAI result length does not match input length"
    )


def test_mrai(prec: Series) -> None:
    precrs = prec.resample("MS").sum()
    mrai_result = mrai(precrs)
    assert isinstance(mrai_result, Series), "MRAI result is not a pandas Series"
    assert len(mrai_result) == len(precrs), (
        "MRAI result length does not match input length"
    )



import logging

import pytest
from pandas import DataFrame, DatetimeIndex, Index, Series, Timestamp, to_datetime

from spei.utils import infer_frequency, validate_index, validate_series


def test_validate_index(caplog) -> None:
    caplog.set_level(logging.INFO)
    series = Series([1.0, 2.0, 3.0], index=["2018", "2019", "2020"])
    validate_index(series.index)
    msg = (
        f"Expected the index to be a DatetimeIndex. Automatically converted "
        f"{type(series.index)} using pd.to_datetime(Index)\n"
    )
    assert msg in caplog.text


def test_validate_index_duplicated(caplog) -> None:
    caplog.set_level(logging.ERROR)
    series = Series(
        [1.0, 1.0],
        index=DatetimeIndex([Timestamp("2000-01-01"), Timestamp("2000-01-01")]),
    )
    with pytest.raises(ValueError):
        validate_index(series.index)
        msg = (
            "Duplicated indices found. Please remove them. For instance by using"
            "`series = series.loc[~series.index.duplicated(keep='first/last')]`"
        )
        assert msg in caplog.text


def test_validate_series() -> None:
    with pytest.raises(TypeError):
        validate_series([1, 2, 3])


def test_validate_series_df_1d(caplog) -> None:
    df = DataFrame({"s": [1, 2, 3]}, index=to_datetime([1, 2, 3]))
    validate_series(df)
    msg = (
        "Please convert series of type pandas.DataFrame to a"
        "pandas.Series using DataFrame.squeeze(). Now done automatically.\n"
    )
    assert msg in caplog.text


def test_validate_series_df_2d() -> None:
    with pytest.raises(TypeError):
        df = DataFrame({"s1": [1, 2, 3], "s2": [1, 2, 3]}, index=to_datetime([1, 2, 3]))
        validate_series(df)


def test_infer_frequency_monthly_start():
    index = DatetimeIndex(["2020-01-01", "2020-02-01", "2020-03-01"])
    assert infer_frequency(index) == "MS"  # Assuming pandas version >= 2.2.0


def test_infer_frequency_monthly_end():
    index = DatetimeIndex(["2020-01-31", "2020-02-28", "2020-03-31"])
    assert infer_frequency(index) == "BME"  # Assuming pandas version >= 2.2.0


def test_infer_frequency_weekly():
    index = DatetimeIndex(["2020-01-01", "2020-01-08", "2020-01-15"])
    assert infer_frequency(index) == "W"


def test_infer_frequency_daily():
    index = DatetimeIndex(["2020-01-01", "2020-01-02", "2020-01-03"])
    assert infer_frequency(index) == "D"


def test_infer_frequency_no_infer():
    index = DatetimeIndex(["2020-01-01", "2020-01-03", "2020-01-07"])
    assert infer_frequency(index) == "MS"  # Assuming pandas version >= 2.2.0


def test_infer_frequency_non_datetime_index():
    index = Index(["2020-01-01", "2020-02-01", "2020-03-01"])
    assert infer_frequency(index) == "MS"  # Assuming pandas version >= 2.2.0


def test_infer_frequency_invalid_index():
    index = Index(["a", "b", "c"])
    with pytest.raises(ValueError):
        infer_frequency(index)



import numpy as np
import pandas as pd
import pytest

from spei.knmi import (
    deficit_apr1,
    deficit_gdd,
    deficit_max,
    deficit_oct1,
    deficit_wet,
    get_cumulative_deficit,
    get_yearly_temp_date,
)
from spei.plot import deficit_knmi


@pytest.fixture
def temp(deficit: pd.Series) -> pd.Series:
    sine_wave = np.sin(2 * np.pi * np.arange(len(deficit)) / 365) * 15 + 15
    temp = pd.Series(data=sine_wave, index=deficit.index, dtype=float)
    return temp


def test_get_yearly_temp_date(temp):
    threshold = 440.0
    result = get_yearly_temp_date(temp=temp, threshold=threshold)
    assert isinstance(result, pd.Series)


def test_get_cumulative_deficit(deficit):
    startdate = pd.Timestamp("2000-04-01")
    enddate = pd.Timestamp("2000-09-30")
    result = get_cumulative_deficit(
        deficit=deficit, startdate=startdate, enddate=enddate
    )
    assert isinstance(result, pd.DataFrame)
    assert not result.empty


def test_deficit_oct1(deficit):
    result = deficit_oct1(deficit=deficit)
    assert isinstance(result, pd.Series)
    assert result.name == "Doct1"


def test_deficit_max(deficit):
    result = deficit_max(deficit=deficit)
    assert isinstance(result, pd.Series)
    assert result.name == "Dmax"


def test_deficit_apr1(deficit):
    result = deficit_apr1(deficit=deficit)
    assert isinstance(result, pd.Series)
    assert result.name == "DIapr1"


def test_deficit_gdd(deficit, temp):
    threshold = 440
    result = deficit_gdd(
        deficit=deficit,
        temp=temp,
        threshold=threshold,
    )
    assert isinstance(result, pd.Series)
    assert result.name == "DIgdd"


def test_deficit_wet(deficit):
    result = deficit_wet(deficit=deficit)
    assert isinstance(result, pd.Series)
    assert result.name == "DIwet"


def test_plot_knmi_deficit(deficit: pd.Series):
    """Test the plot function for the deficit."""
    startdate = pd.Timestamp("2000-04-01")
    enddate = pd.Timestamp("2000-09-30")
    cumdf = get_cumulative_deficit(
        deficit=deficit,
        startdate=startdate,
        enddate=enddate,
        allow_below_zero=False,
    )
    ax = deficit_knmi(cumdf)
    assert ax is not None
    assert ax.get_ylabel() == "Precipitation deficit (mm)"



import logging
from calendar import isleap

from numpy import array, nan
from pandas import (
    DataFrame,
    DatetimeIndex,
    Grouper,
    Index,
    Series,
    Timedelta,
    concat,
    infer_freq,
    to_datetime,
)
from pandas import __version__ as pd_version


def validate_series(series: Series) -> Series:
    series = series.copy()

    if not isinstance(series, Series):
        if isinstance(series, DataFrame):
            if len(series.columns) == 1:
                logging.warning(
                    "Please convert series of type pandas.DataFrame to a"
                    "pandas.Series using DataFrame.squeeze(). Now done automatically."
                )
                series = series.squeeze()
            else:
                raise TypeError(
                    "Please provide a pandas.Series instead of a pandas.DataFrame"
                )
        else:
            raise TypeError(f"Please provide a Pandas Series instead of {type(series)}")

    index = validate_index(series.index)

    return series.reindex(index, copy=True)


def validate_index(index: Index) -> DatetimeIndex:
    index = index.copy()

    if not isinstance(index, DatetimeIndex):
        logging.info(
            f"Expected the index to be a DatetimeIndex. Automatically converted "
            f"{type(index)} using pd.to_datetime(Index)"
        )
        index = DatetimeIndex(to_datetime(index))

    if index.has_duplicates:
        msg = (
            "Duplicated indices found. Please remove them. For instance by"
            " using `series = "
            "series.loc[~series.index.duplicated(keep='first/last')]`"
        )
        logging.error(msg)
        raise ValueError(msg)

    return index


def infer_frequency(index: Index | DatetimeIndex) -> str:
    """Infer frequency"""

    index = validate_index(index)

    inf_freq = infer_freq(index)

    if inf_freq is None:
        logging.info(
            "Could not infer frequency from index, using monthly frequency instead"
        )
        inf_freq = "MS" if pd_version >= "2.2.0" else "M"
    else:
        logging.info(f"Inferred frequency '{inf_freq}' from index")

    if "W-" in inf_freq:
        logging.info(f"Converted frequncy weekly '{inf_freq}' to 'W'")
        inf_freq = "W"

    return inf_freq


def group_yearly_df(series: Series) -> DataFrame:
    """Group series in a DataFrame with date (in the year 2000) as index and
    year as columns.
    """
    strfstr: str = "%m-%d %H:%M:%S"
    grs = {}
    freq = "YE" if pd_version >= "2.2.0" else "Y"
    for year_timestamp, gry in series.groupby(Grouper(freq=freq)):
        index = validate_index(gry.index)
        gry.index = to_datetime(
            "2000-" + index.strftime(strfstr), format="%Y-" + strfstr
        )
        year = getattr(year_timestamp, "year")  # type: str
        grs[year] = gry
    return concat(grs, axis=1)


def get_data_series(group_df: DataFrame) -> Series:
    """Transform grouped dataframe by yearly values back to time series."""
    strfstr: str = "%m-%d %H:%M:%S"
    index = validate_index(group_df.index)
    idx = array(
        [(f"{col}-" + index.strftime(strfstr)).tolist() for col in group_df.columns]
    ).flatten()
    # remove illegal 29 february for non leap years created by group_yearly_df
    boolidx = ~array(
        [
            (x.split(" ")[0].split("-", 1)[1] == "02-29")
            and not isleap(int(x.split(" ")[0].split("-")[0]))
            for x in idx
        ]
    )

    dt_idx = to_datetime(idx[boolidx], format="%Y-" + strfstr)
    values = group_df.transpose().values.flatten()[boolidx]
    return Series(values, index=dt_idx, dtype=float).dropna()


def daily_window_group_yearly_df(dfval: DataFrame, period: int) -> DataFrame:
    """Fill a period of daily values in grouped by yearly DataFrame to get
    cyclic rolling window.
    """
    dfval_window_index_start = [
        dfval.index[0] + Timedelta(value=-i, unit="D")
        for i in reversed(range(1, period + 1))
    ]
    dfval_window_index_end = [
        dfval.index[-1] + Timedelta(value=i, unit="D") for i in range(1, period + 1)
    ]
    dfval_window_index = DatetimeIndex(
        dfval_window_index_start + dfval.index.to_list() + dfval_window_index_end
    )

    dfval_window = DataFrame(
        nan, index=dfval_window_index, columns=dfval.columns, dtype=float
    )
    dfval_window.loc[dfval.index, dfval.columns] = dfval.values
    dfval_window.iloc[:period] = dfval.iloc[-period:].values
    dfval_window.iloc[-period:] = dfval.iloc[:period].values
    return dfval_window



# flake8: noqa
from typing import Any

from numpy import float64
from numpy.typing import NDArray
from scipy.stats._continuous_distns import rv_continuous

ContinuousDist = Any | rv_continuous
NDArrayAxes = NDArray[Any]
NDArrayFloat = NDArray[float64]



from dataclasses import dataclass, field
from typing import Literal

from numpy import std
from pandas import Series
from scipy.stats import kstest

from ._typing import ContinuousDist


@dataclass
class Dist:
    data: Series = field(init=True, repr=False)
    dist: ContinuousDist
    loc: float = field(init=False, repr=True)
    scale: float = field(init=False, repr=True)
    pars: list[float] | None = field(init=False, repr=False)
    prob_zero: bool = field(default=False, init=True, repr=False)
    p0: float = field(default=0.0, init=False, repr=False)
    data_window: Series | None = field(default=None, init=True, repr=False)
    """
    Represents a distribution associated with data.

    Parameters
    ----------
    data : Series
        The input data for fitting the distribution.
    dist : ContinuousDist
        The SciPy continuous distribution associated to be fitted.
    prob_zero : bool, default=False
        Flag indicating whether the probability of zero values in the series is
        calculated by the occurence.
    data_window : Optional[Series], default=None
        Subset of data for fitting more data (if provided).

    Attributes
    ----------
    loc : float
        Location of the distribution
    scale : float
        Scale of the distribution
    pars : Optional[List[float]]
        Attribute storing additional distribution parameters (if applicable).
    p0 : float
        The probability of zero values in the data. Only calculated if prob_zero=True.

    Methods
    -------
    __post_init__(self) -> None
        Initializes the Dist class and fits the distribution.
    fit_dist(data: Series, dist: ContinuousDist) -> Tuple
        Fits a Scipy continuous distribution to the data.

    Notes
    -----
    The `fit_dist` method uses the `dist.fit` function from Scipy to estimate
    distribution parameters. If the fitted distribution requires additional
    parameters beyond `loc` and `scale`, they are stored in the `pars` attribute.
    """

    def __post_init__(self):
        """
        Post initializes the Dist class by fitting the distribution.
        """
        data_fit = self.data_window if self.data_window is not None else self.data
        pars, loc, scale = self.fit_dist(data=data_fit, dist=self.dist)
        self.loc = loc
        self.scale = scale
        self.pars = pars

        if self.prob_zero:
            self.p0 = (data_fit == 0.0).sum() / len(data_fit)

    @staticmethod
    def fit_dist(
        data: Series, dist: ContinuousDist
    ) -> tuple[list[float] | None, float, float]:
        """
        Fits a Scipy continuous distribution to the data.

        Parameters
        ----------
        data : Series
            The input data for fitting.
        dist : ContinuousDist
            The continuous distribution to be fitted.

        Returns
        -------
        Tuple
            Tuple containing distribution parameters (pars, loc, scale).
        """
        fit_tuple = dist.fit(data, scale=std(data))
        if len(fit_tuple) == 2:
            loc, scale = fit_tuple
            pars = None
        else:
            *pars, loc, scale = fit_tuple
        return pars, loc, scale

    def cdf(self) -> Series:
        """Compute cumulative density function of a Scipy Continuous Distribution"""
        if self.pars is not None:
            cdf = self.dist.cdf(
                self.data.values, *self.pars, loc=self.loc, scale=self.scale
            )
        else:
            cdf = self.dist.cdf(self.data.values, loc=self.loc, scale=self.scale)

        if self.prob_zero:
            cdf = self.p0 + (1 - self.p0) * cdf
            cdf[self.data == 0.0] = self.p0

        return Series(cdf, index=self.data.index, dtype=float)

    def pdf(self) -> Series:
        data_pdf = self.data.sort_values()
        if self.pars is not None:
            pdf = self.dist.pdf(
                data_pdf.values, *self.pars, loc=self.loc, scale=self.scale
            )
        else:
            pdf = self.dist.pdf(data_pdf.values, loc=self.loc, scale=self.scale)

        if self.prob_zero:
            pdf = self.p0 + (1 - self.p0) * pdf
            pdf[self.data == 0.0] = self.p0

        return Series(pdf, index=data_pdf.index, dtype=float)

    def ppf(self, q: float) -> Series:
        """Compute percent point function (inverse of cdf) at q"""
        if self.pars is not None:
            ppf = self.dist.ppf(q, *self.pars, loc=self.loc, scale=self.scale)
        else:
            ppf = self.dist.ppf(q, loc=self.loc, scale=self.scale)

        return Series(ppf, index=self.data.index, dtype=float)

    def ks_test(
        self,
        method: Literal["auto", "exact", "approx", "asymp"] = "auto",
    ) -> float:
        """Fit a distribution and perform the two-sided
        Kolmogorov-Smirnov test for goodness of fit. The
        null hypothesis is that the data and distributions
        are identical, the alternative is that they are
        not identical. [scipy_2021]_

        Parameters
        ----------
        method : Literal['auto', 'exact', 'approx', 'asymp'], optional
            Defines the distribution used for calculating the p-value. The
            following options are available (default is 'auto'): 'auto' selects
            one of the other options, 'exact' uses the exact distribution of
            test statistic, 'approx' approximates the two-sided probability
            with twice the one-sided probability, 'asymp' uses asymptotic
            distribution of test statistic

        Returns
        -------
        float
            p-value

        References
        -------
        .. [scipy_2021] Onnen, H.: Intro to Probability
        Distributions and Distribution Fitting with Pythons
        SciPy, 2021.
        """
        args = (
            (self.pars, self.loc, self.scale)
            if self.pars is not None
            else (self.loc, self.scale)
        )
        kstest_result = kstest(
            rvs=self.data, cdf=self.dist.name, args=args, method=method
        )
        # rej_h0 = kstest_result.pvalue < alpha
        return kstest_result.pvalue



import spei as si


def test_version() -> None:
    assert isinstance(si.__version__, str)
    assert si.__version__.count(".") == 2


def test_show_versions():
    msg = si.show_versions()
    assert isinstance(msg, str)
